/*
	gb2fasta.c

	Convert a list of GenBank-format sequence files into FASTA format
	(for subsequent processing by PRESSDB and use by BLASTN).

	Usage:  gb2fasta gbpri.seq gbrod.seq gbmam.seq ... > genbank.total

	Suggested alternative usage:

	gb2fasta -l GenPept genpept.seq > genpept

	Tested successfully on GenBank Release 66.0

	7/22/90 WRG
*/
#include <ncbi.h>
#include <gishlib.h>

FILE	*fp;
#define BUFFER_SIZE	256
char	buf[BUFFER_SIZE];

Boolean	label_set;
char	*label = "gb";
char	*label_default = "gb";

char	*genbank_letters = "ISTGHUJKLMN";
char	*embl_letters = "VXWYZA";
char	*ddbj_letters = "DE";
char	*swissprot_letters = "PQ";

char	*module;
char	locus[128];
char	accession[128];
char	definition[4096];
Boolean	gflag = TRUE;
unsigned long	giid;

main(ac, av)
	int		ac;
	char	**av;
{
	FILE	*fdopen();
	int		i, len, len2;
	register char	*cp, ch;

	module = str_dup(basename(av[0], NULL));

	if (ac == 2 && strcmp(av[1], "-") == 0)
		optind = ac-1;
	else
		while ((i = getopt(ac, av, "gl:")) != -1)
			switch (i) {
			case 'g':
				gflag = FALSE;
				break;
			case 'l':
				label = optarg;
				label_set = TRUE;
				break;
			default:
				usage();
			}

	if (ac == optind)
		usage();

	for (i = optind; i < ac; ++i) {
		fp = openfile(av[i], "rb");
		if (fp == NULL)
			goto NextFile;

		for (;/* Each Nt. Sequence */;) {
			locus[0] = accession[0] = definition[0] = NULLB;
			for (;/* Each Record */;) {
				if (fgets(buf, BUFFER_SIZE, fp) == NULL)
					goto NextFile;
Switch:
				switch (buf[0]) {
				case 'L':
					giid = 0;
					accession[0] = NULLB;
					definition[0] = NULLB;
					if (strncmp(buf, "LOCUS", 5))
						continue;
					get_locus(buf);
					continue;
				case 'D':
					if (strncmp(buf, "DEFINITION", 10))
						continue;
					get_definition(buf);
					goto Switch;
				case 'A':
					if (strncmp(buf, "ACCESSION", 9))
						continue;
					if (accession[0] != NULLB)
						continue;
					if (strtok(buf, " \t\n\r") != NULL) {
						cp = strtok(NULL, " \t\n\r");
						if (cp != NULL)
							strcpy(accession, cp);
					}
					continue;
				case 'O':
					if (strncmp(buf, "ORIGIN", 6) == 0)
						goto GetSeq;
					continue;
				case 'N':
					/* this was an old way of including the gi */
					if (gflag && strncmp(buf, "NCBI_GI", 7) == 0) {
						sscanf(buf + 8, "%lu", &giid);
					}
					continue;
				case 'C':
					if (!gflag)
						continue;
					if (giid == 0 && strncmp(buf, "COMMENT", 7) == 0)
						get_gi(buf);
					goto Switch;
				default:
					continue;
				}
			}

GetSeq:
			putchar('>');
			if (giid > 0) {
				printf("gi|%lu|", giid);
			}
			if (!label_set) {
				label = label_default;
				if (accession[0] != NULLB) {
					if (strchr(genbank_letters, accession[0]) != NULL)
						label = "gb";
					else
						if (strchr(embl_letters, accession[0]) != NULL)
							label = "emb";
						else
							if (strchr(ddbj_letters, accession[0]) != NULL)
								label = "dbj";
				}
			}
			if (label[0] != NULLB)
				fputs(label, stdout);
			else
				fputs("???", stdout);
			putchar('|');
			if (accession[0] != NULLB)
				fputs(accession, stdout);
			putchar('|');
			if (locus[0] != NULLB)
				fputs(locus, stdout);

			if (definition[0]) {
				putchar(' ');
				fputs(definition, stdout);
			}
			putchar('\n');

			while (fgets(buf, BUFFER_SIZE, fp) != NULL && buf[0] != '/') {
				cp = buf;
				while ((ch = *cp++) != NULLB)
					if (isalpha(ch)) {
						ch = TOUPPER(ch);
						putchar(ch);
					}
				putchar('\n');
			}

		} /* for (;;) */
		

NextFile:
		if (fp != NULL)
			fclose(fp);
		continue;
	}
	exit(0);
}

int
get_gi(buf)
	register CharPtr	buf;
{
	if (giid > 0)
		return 0;
	buf[0] = ' ';
	do {
		if (buf[0] != ' ' && buf[0] != '\n')
			return 0;
		if (buf[12] == 'N' && strncmp(buf+12, "NCBI gi:", 8) == 0) {
			sscanf(buf+20, "%lu", &giid);
			return 0;
		}
	} while (fgets(buf, BUFFER_SIZE, fp) != NULL);
	return 1;
}

int
get_locus(buf)
	register CharPtr	buf;
{
	register CharPtr	cp;
	register int	ch, len;

	if (locus[0] != NULLB) {
		fprintf(stderr, "\nLOCUS without a sequence\n");
		fprintf(stderr, "LOCUS %s\n", locus);
		exit(1);
	}
	if ((cp = strtok(buf+6, " \t\n\r")) != NULL)
		strcpy(locus, cp);
	else {
		fprintf(stderr, "\nNo locus found on LOCUS line\n");
		exit(1);
	}
	return 0;
}

int
get_definition(buf)
	register CharPtr	buf;
{
	register CharPtr	cp;
	register int		ch, len, len2;

	if (definition[0] != NULLB) {
		fprintf(stderr, "\nDuplicate DEFINITION records for a single sequence\n");
		if (locus[0])
			fprintf(stderr, "LOCUS %s\n", locus);
		exit(2);
	}

	cp = buf+10;
	while ((ch = *cp) != NULLB && isspace(ch))
		++cp;
	definition[len = 0] = NULLB;
	if (ch != NULLB) {
		len = strlen(cp) - 1;
		MemCpy(definition, cp, len);
		definition[len] = NULLB;
		if (len > 0 && definition[len-1] == '\r') /* CD-ROM? */
			definition[--len] = NULLB;
	}
	while (fgets(buf, BUFFER_SIZE, fp) != NULL &&
				buf[0] == ' ') {
		cp = buf+1;
		while ((ch = *cp) != NULLB && isspace(ch))
			++cp;
		if (ch != NULLB) {
			len2 = strlen(cp) - 1;
			definition[len++] = ' ';
			MemCpy(definition + len, cp, len2);
			len += len2;
			definition[len] = NULLB;
			if (len > 0 && definition[len-1] == '\r') /* CD-ROM? */
				definition[--len] = NULLB;
		}
	}
	if (len == 1 && definition[0] == '.')
		strcpy(definition, "No DEFINITION available");
	return 0;
}


usage()
{
	fprintf(stderr, "Purpose:  convert a GenBank-format file into FASTA format.\n");
	fprintf(stderr, "Usage:  %s [-g] [-l label] dbfile\n", module);
	exit(1);
}
