/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <gishlib.h>
#include <signal.h>

static PoolBlkPtr	apool;
static AlarmBlkPtr	a0, a1;
static sigset_t	vec;

static void	alarm_reset PROTO((AlarmBlkPtr, time_t));
static void	alarm_remove PROTO((AlarmBlkPtr));
static void	alarm_add PROTO((AlarmBlkPtr));
static AlarmBlkPtr alarm_set PROTO((time_t when, FnPtr func, VoidPtr arg));

static void
alarm_handler(sig)
	int	sig;
{
	AlarmBlk	ab;
	register AlarmBlkPtr	ap;
	register time_t	now, t;

	if (a0 == NULL)
		return;

	now = time(NULL);
	while (a0->dingdong <= now) {
		ab.func = a0->func;
		ab.arg = a0->arg;
		if (a0->period == 0)
			alarm_remove(a0);
		else
			alarm_reset(a0, a0->period);
		(*ab.func)(ab.arg);
	}

	if (a0 != NULL) {
		signal(SIGALRM, alarm_handler);
		now = time(NULL);
		if (a0->dingdong <= now)
			t = 1;
		else
			t = a0->dingdong - now;
		alarm(t);
	}
}

/* alarm_reset -- set a new alarm time */
static void
alarm_reset(abp, when)
	register AlarmBlkPtr	abp;
	time_t	when;
{
	register AlarmBlkPtr	ap, ap2;

	if (when == 0)
		return;

	/* Find the previous pointer to abp */
	for (ap2 = NULL, ap = a0; ap != NULL && ap != abp; ap2 = ap, ap = ap->next)
		;
	if (ap == NULL)
		return; /* not found */
	/* Remove item from linked list */
	if (ap2 != NULL)
		ap2->next = ap->next;
	else
		a0 = ap->next;

	abp->dingdong = when + time(NULL);
	alarm_add(abp);
	return;
}

static void
alarm_remove(abp)
	AlarmBlkPtr	abp;
{
	register AlarmBlkPtr	ap, ap2;

	if (abp == NULL) { /* remove all alarms */
		alarm(0);
		PoolPutLink(apool, (VoidPtr)a0);
		return;
	}

	for (ap2 = NULL, ap = a0; ap != NULL; ap2 = ap, ap = ap->next) {
		if (ap == abp) {
			if (ap2 == NULL)
				a0 = ap->next;
			else
				ap2->next = ap->next;
			PoolPut(apool, (VoidPtr) ap);
			return;
		}
	}
	/* item not found */
}

/* alarm_add -- insert an event in temporal order in the linked list */
static void
alarm_add(abp)
	AlarmBlkPtr	abp;
{
	register AlarmBlkPtr	ap, ap2;
	register time_t	t, now;

	for (ap2 = NULL, ap = a0;
			ap != NULL && ap->dingdong < abp->dingdong;
			ap2 = ap, ap = ap->next) {
	}
	if (ap2 == NULL) {
		a0 = abp;
		abp->next = ap;
		signal(SIGALRM, alarm_handler);
		now = time(NULL);
		if (abp->dingdong <= now)
			t = 1;
		else
			t = abp->dingdong - now;
		alarm(t);
	}
	else {
		ap2->next = abp;
		abp->next = ap;
	}
}

static AlarmBlkPtr
alarm_set(when, func, arg)
	time_t	when;
	FnPtr	func;
	VoidPtr	arg;
{
	register AlarmBlkPtr	abp;

	if (apool == NULL) {
		apool = PoolNew(20, sizeof(AlarmBlk), offsetof(AlarmBlk,next), NULL);
		if (apool == NULL)
			return NULL;
	}

	abp = (AlarmBlkPtr) PoolGet(apool);
	if (abp != NULL) {
		abp->next = NULL;
		abp->period = 0;
		abp->dingdong = when + time(NULL);
		abp->func = func;
		abp->arg = arg;
		alarm_add(abp);
	}

	return abp;
}

AlarmBlkPtr
AlarmSet(when, func, arg)
	time_t	when;
	FnPtr	func;
	VoidPtr	arg;
{
	AlarmBlkPtr	ap;

	if (when == 0)
		return NULL;
	critical_begin();
	ap = alarm_set(when, func, arg);
	critical_end();
	return ap;
}


AlarmBlkPtr
AlarmEvery(period, func, arg)
	time_t	period;
	FnPtr	func;
	VoidPtr	arg;
{
	AlarmBlkPtr	ap;

	if (period == 0)
		return NULL;
	critical_begin();
	ap = alarm_set(period, func, arg);
	if (ap != NULL)
		ap->period = period;
	critical_end();
	return ap;
}

void
AlarmReset(abp, when)
	AlarmBlkPtr	abp;
	time_t	when;
{
	critical_begin();
	alarm_reset(abp, when);
	critical_end();
}

void
AlarmClr(abp)
	AlarmBlkPtr	abp;
{
	if (abp == NULL)
		return;

	critical_begin();
	alarm_remove(abp);
	critical_end();
}

void
AlarmClrAll()
{
	critical_begin();
	alarm_remove(NULL);
	critical_end();
}

time_t
AlarmWhen()
{
	register time_t	t, now;

	critical_begin();
	t = (a0 != NULL ? a0->dingdong : 0);
	critical_end();
	if (t == 0)
		return t;

	now = time(NULL);
	if (t < now)
		return 0;
	return t - now;
}
